from typing import Dict
import numpy as np
import pandas as pd
from tqdm.notebook import tqdm

from .signal import get_signal, get_single_signal
from .phi import phi_ar, calc_phi_for_signal
from .utils import fill_nans_with_mean, sliding_window

def phi_for_act_dict(
    acts : Dict[str, list], 
    time_scale : int = 24*3600,
    window : int = 30, 
    hop : int = 10,
    binarize : bool = False,
    base : float = np.e,
    filter_resol : int = 20,
    nan_policy : str = 'zero',
    silent : bool = False):
    '''
    ** parameters **

    acts = a dictionary with keys equal to user_id and values equal to a list of timebins in which the corresponding user was active
    time_scale = in seconds. used as the scale of time for future calculations
    window = window length used for distribution estimation
    hop = time delay for phi calculation
    binarize = whether to binarize the activity signal (instead of number of activites per timebin use 1 as active and 0 as inactive)
    base = base of the logarithms used in calculating phi. base=e means phi is in nats and base=2 means phi is in bits.
    nan_policy = {
        'zero' : fill nans with zeros
        'none' : do nothing
        'linear' : fill nans with linear interpolation
    }
    silent = disable progress bar



    ** outputs **
    phis = list of phi values for each time step
    n_users = list of int indicating number of users used for calculating phi in the k-th timestep.

    '''
    
    sig = get_signal(acts, time_scale=time_scale, binarize=binarize)
    return phi_for_act_sig(
        sig, base=base, window=window, hop=hop, filter_resol=filter_resol, nan_policy=nan_policy, silent=silent
    )


def phi_for_act_sig(
    sig : np.ndarray, 
    window : int = 30, 
    hop : int = 10,
    base : float = np.e, 
    filter_resol : int = 20,
    nan_policy : str = 'zero',
    silent : bool = False):
    '''
    ** parameters **

    sig = a 2-D numpy array with shape=(n_user, n_timesteps)
    window = window length used for distribution estimation
    hop = time delay for phi calculation
    base = base of the logarithms used in calculating phi. base=e means phi is in nats and base=2 means phi is in bits.
    nan_policy = {
        'zero' : fill nans with zeros
        'none' : do nothing
        'linear' : fill nans with linear interpolation
    }
    silent = disable progress bar
    

    ** outputs **
    phis = list of phi values for each time step
    n_users = list of int indicating number of users used for calculating phi in the k-th timestep.

    '''

    phis, n_users = calc_phi_for_signal(sig, win_len=window, hop_len=hop, base=base, filter_resol=filter_resol, silent=silent)
    if nan_policy == 'zero':
        phis[np.isnan(phis)] = 0.
    elif nan_policy == 'linear':
        phis = fill_nans_with_mean(phis)
    return phis, n_users

def experiment_hop_range(
    sig : np.ndarray, 
    window : int = 30, 
    min_hop : int = 1,
    max_hop : int = 7, 
    filter_resol : int = 20,
    silent : bool = False):

    res = {}
    prog = lambda x: x if silent else tqdm(x)
    for hop in prog(range(min_hop, max_hop+1)):
        phis, _ = calc_phi_for_signal(sig, win_len=window, hop_len=hop, base=2, filter_resol=filter_resol, silent=True)
        nans = np.isnan(phis).sum()
        zero_phis = phis.copy()
        zero_phis[np.isnan(zero_phis)] = 0.
        
        res[hop] = {
            'hop' : hop,
            'num_nans' : nans, 
            'num_valids': len(phis) - nans, 
            'vnr': (len(phis) - nans)/nans, 
            'raw_avg_phi': np.mean(phis[~np.isnan(phis)]),
            'zeroed_avg_phi': np.mean(zero_phis),
            'linear_avg_phi' : np.mean(fill_nans_with_mean(phis))
        }
    return pd.DataFrame(res).T



import numpy as np
from tqdm.notebook import tqdm

def sliding_window(arr, agg_func=np.mean, window=10, hop=10):
    res = []
    n = (arr.shape[-1] - window) // hop
    for i in range(n):
        res += [agg_func(arr[..., i*hop : i*hop + window], axis=1)]
    return np.stack(res, axis=1)

def nan_mean_value(t_arr, nan_idx):
    arr = t_arr.copy()
    idx1 = idx2 = None
    if np.isnan(arr[0]):
        arr = np.append(0, arr)
        nan_idx += 1
    if np.isnan(arr[-1]):
        arr = np.append(arr, 0)
    for i in range(nan_idx-1,-1,-1):
        if not np.isnan(arr[i]):
            idx1 = i
            break
            
    for i in range(nan_idx+1, len(arr)):
        if not np.isnan(arr[i]):
            idx2 = i
            break 

    m = (arr[idx2] - arr[idx1]) / (idx2 - idx1)
    return arr[idx1] + m*(nan_idx - idx1)

def fill_nans_with_mean(arr):
    arr = arr.copy()
    nans = np.where(np.isnan(arr))[0]
    for nan in nans:
        arr[nan] = nan_mean_value(arr, nan)
    return arr


import numpy as np
from collections import Counter


def get_single_signal(act, time_scale=24*3600, binarize=False):
    act = dict(Counter([int(t // time_scale) for t in act]))
    steps = sorted(act.keys()) #act
    a,b = steps[-1], steps[0]
    sig = np.zeros(shape=(a-b+1,))
    for step in steps:
        v = 1 if binarize else act[step]
        sig[step - b] = v
    return sig, a, b

def get_signal(activities, time_scale=24*3600, binarize=False):
    min_step = 999999999999999
    max_step = 0
    users_acts = {}
    
    for user, act in activities.items():
        sig, M, m = get_single_signal(act, time_scale, binarize)
        min_step = min(m, min_step)
        max_step = max(M, max_step)
        users_acts[user] = {'min': m, 'max': M, 'sig': sig}

    for user in users_acts:
        users_acts[user]['sig'] = np.pad(users_acts[user]['sig'], 
                                         pad_width=(users_acts[user]['min'] - min_step, 
                                                    max_step - users_acts[user]['max']))

    return np.stack([users_acts[user]['sig'] for user in users_acts], axis = 0)


from setuptools import setup

url = 'https://github.com/s-omranpour/Social-Phi'
setup(
    name='social_phi',
    version='0.0.1',
    description='a python package to calculate temporal integrated information of a group of people based on their temporal activites as time series',
    url=url,
    author='Soroush Omranpour',
    author_email='soroush.333@gmail.com',
    keywords='information theory, integrated information, consciousness, social network',
    packages=['social_phi'],
    python_requires='>=3.7, <4',
    install_requires=[
        'numpy >= 1.7.0'
    ],
    license="MIT license",

    project_urls={  # Optional
        'Bug Reports': url + '/issues',
        'Source': url,
    },
)


import numpy as np
from tqdm.notebook import tqdm

def regression(X, Y):
    X_p = np.concatenate([np.ones(shape=(X.shape[0],1)), X], axis=1)
    beta, _, _ , _ = np.linalg.lstsq(X_p, Y, rcond=None)
    E = Y - X_p@beta
    return E


def phi_ar(X1, X2, base=np.e):
    assert X1.shape == X2.shape
    prod = lambda x: x[0]*x[1]
    
    S_X1 = np.cov(X1)
    S_part_X1_X2 = np.cov(regression(X2, X1))
    numerator = prod(np.linalg.slogdet(S_X1))
    denominator = prod(np.linalg.slogdet(S_part_X1_X2))
    term1 = 0.5*(numerator-denominator)

    S_M1 = [S_X1[i,i] for i in range(X1.shape[0])]
    S_part_M1_M2 = [np.cov(regression(X2[i:i+1], X1[i:i+1])) for i in range(X1.shape[0])]
    term2 = 0.5*np.sum([np.log(s_m1) - np.log(s_part_m1_m2) for s_m1, s_part_m1_m2 in zip(S_M1, S_part_M1_M2)])

    return (term1 - term2)/ np.log(base)


def calc_phi_for_signal(sig, win_len=100, hop_len=1, base=np.e, filter_resol=20, silent=False):
    phis = []
    num_users = []
    X = sig
    n = (X.shape[1] - win_len) // hop_len 
    prog = lambda x: x if silent else tqdm(x)
    for i in prog(range(n)):
        X1 = X[:, i*hop_len : i*hop_len + win_len]
        X2 = X[:, (i+1)*hop_len : (i+1)*hop_len + win_len]
        phi = np.nan
        
        ## filtering zero var users
        valid_users = np.var(X1, axis=1) > 0
        X1 = X1[valid_users]
        X2 = X2[valid_users]
        
        m = X1.shape[0]
        n_user = m
        users_sorted = np.argsort(np.var(X1, axis=1))
        for i in range(0,m,max(1, m//filter_resol)):
            top_users = users_sorted[i:]
            n_user = len(top_users)
            if n_user < 2:
                continue

            res = phi_ar(X1[top_users], X2[top_users], base=base)
            if res >= 0 and res <= n_user:
                phi = res
                break

        phis += [phi]
        num_users += [n_user]
    return np.array(phis), np.array(num_users)
